/* Server.java
 *
 * Copyright (C) 2006-2013 wolfSSL Inc.
 *
 * This file is part of CyaSSL.
 *
 * CyaSSL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CyaSSL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

import java.io.*;
import java.net.*;
import java.nio.*;
import java.nio.charset.Charset;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CharacterCodingException;

import com.wolfssl.WolfSSL;
import com.wolfssl.WolfSSLSession;
import com.wolfssl.WolfSSLContext;
import com.wolfssl.WolfSSLException;
import com.wolfssl.WolfSSLIOSendCallback;
import com.wolfssl.WolfSSLIORecvCallback;

public class Server {
    
    public static Charset charset = Charset.forName("UTF-8");
    public static CharsetEncoder encoder = charset.newEncoder();

    public static void main(String[] args) {
        new Server().run(args);
    }

    public void run(String[] args) {

        Socket clientSocket           = null;
        ServerSocket serverSocket     = null;
        DatagramSocket d_serverSocket = null;
        DataOutputStream outstream    = null;
        DataInputStream  instream     = null;

        int ret, insz;
        String msg  = "I hear you fa shizzle, from Java!";
        byte[] input = new byte[80];
        long method = 0;

        /* config info */
        boolean useIOCallbacks = false;       /* test I/O callbacks */
        String cipherList = "AES128-SHA";     /* try AES128-SHA by default */
        int sslVersion = 3;                   /* default to TLS 1.2 */
        int verifyPeer = 1;                   /* verify peer by default */
        int doDTLS = 0;                       /* don't use DTLS by default */
        int useOcsp = 0;                      /* don't use OCSP by default */
        String ocspUrl = null;                /* OCSP override URL */
        int useAtomic = 0;                    /* atomic record lyr processing */
        int pkCallbacks = 0;                  /* public key callbacks */
        int logCallback = 0;                  /* test logging callback */

        /* cert info */
        String serverCert = "../certs/server-cert.pem";
        String serverKey  = "../certs/server-key.pem";
        String caCert     = "../certs/client-cert.pem";
        String crlPemDir  = "../certs/crl";

        /* server info */ 
        int port    =  11111;

        /* pull in command line options from user */
        for (int i = 0; i < args.length; i++)
        {
            String arg = args[i];

            if (arg.equals("-?")) {
                printUsage();

            } else if (arg.equals("-p")) {
                if (args.length < i+2)
                    printUsage();
                port = Integer.parseInt(args[++i]);

            } else if (arg.equals("-v")) {
                if (args.length < i+2)
                    printUsage();
                sslVersion = Integer.parseInt(args[++i]);
                if (sslVersion < 0 || sslVersion > 3) {
                    printUsage();
                }
    
            } else if (arg.equals("-l")) {
                if (args.length < i+2)
                    printUsage();
                cipherList = args[++i];

            } else if (arg.equals("-c")) {
                if (args.length < i+2)
                    printUsage();
                serverCert = args[++i];

            } else if (arg.equals("-k")) {
                if (args.length < i+2)
                    printUsage();
                serverKey = args[++i];

            } else if (arg.equals("-A")) {
                if (args.length < i+2)
                    printUsage();
                caCert = args[++i];

            } else if (arg.equals("-d")) {
                verifyPeer = 0;

            } else if (arg.equals("-u")) {
                doDTLS = 1;

            } else if (arg.equals("-iocb")) {
                useIOCallbacks = true;

            } else if (arg.equals("-logtest")) {
                logCallback = 1;

            } else if (arg.equals("-o")) {
                useOcsp = 1;

            } else if (arg.equals("-O")) {
                if (args.length < i+2)
                    printUsage();
                useOcsp = 1;
                ocspUrl = args[i++];

            } else if (arg.equals("-U")) {
                useAtomic = 1;

            } else if (arg.equals("-P")) {
                pkCallbacks = 1;
            
            } else {
                printUsage();
            }
        }

        /* sort out DTLS versus TLS versions */
        if (doDTLS == 1) {
            if (sslVersion == 3)
                sslVersion = -2;
            else
                sslVersion = -1;
        }

        try {

            /* load JNI library */
            WolfSSL.loadLibrary();

            /* init library */
            WolfSSL sslLib = new WolfSSL();
            sslLib.debuggingON();

            /* set logging callback */
            if (logCallback == 1) {
                MyLoggingCallback lc = new MyLoggingCallback();
                sslLib.setLoggingCb(lc);
            }

            /* set SSL version method */
            switch (sslVersion) {
                case 0:
                    method = WolfSSL.SSLv3_ServerMethod();
                    break;
                case 1:
                    method = WolfSSL.TLSv1_ServerMethod();
                    break;
                case 2:
                    method = WolfSSL.TLSv1_1_ServerMethod();
                    break;
                case 3:
                    method = WolfSSL.TLSv1_2_ServerMethod();
                    break;
                case -1:
                    method = WolfSSL.DTLSv1_ServerMethod();
                    break;
                case -2:
                    method = WolfSSL.DTLSv1_2_ServerMethod();
                    break;
                default:
                    System.err.println("Bad SSL version");
                    System.exit(1);
            }

            /* create context */
            WolfSSLContext sslCtx = new WolfSSLContext(method);

            /* load certificate/key files */
            ret = sslCtx.useCertificateFile(serverCert,
                    WolfSSL.SSL_FILETYPE_PEM);
            if (ret != WolfSSL.SSL_SUCCESS) {
                System.out.println("failed to load server certificate!");
                System.exit(1);
            }

            ret = sslCtx.usePrivateKeyFile(serverKey,
                    WolfSSL.SSL_FILETYPE_PEM);
            if (ret != WolfSSL.SSL_SUCCESS) {
                System.out.println("failed to load server private key!");
                System.exit(1);
            }

                
            /* set verify callback */
            if (verifyPeer == 0) {
                sslCtx.setVerify(WolfSSL.SSL_VERIFY_NONE, null);
            } else {
                ret = sslCtx.loadVerifyLocations(caCert, null);
                if (ret != WolfSSL.SSL_SUCCESS) {
                    System.out.println("failed to load CA certificates!");
                    System.exit(1);
                }

                VerifyCallback vc = new VerifyCallback();
                sslCtx.setVerify(WolfSSL.SSL_VERIFY_PEER, vc);
            }

            /* set cipher list */
            if (cipherList != null)
                ret = sslCtx.setCipherList(cipherList);

            /* set OCSP options, override URL */
            if (useOcsp == 1) {
                long ocspOptions = WolfSSL.CYASSL_OCSP_ENABLE |
                                   WolfSSL.CYASSL_OCSP_NO_NONCE;
                ret = sslCtx.setOCSPOptions(ocspOptions);
                if (ret != WolfSSL.SSL_SUCCESS) {
                    System.out.println("failed to set OCSP options, ret = "
                            + ret);
                    System.exit(1);
                }

                if (ocspUrl != null) {
                    ret = sslCtx.setOCSPOverrideUrl(ocspUrl);

                    if (ret != WolfSSL.SSL_SUCCESS) {
                        System.out.println("failed to set OCSP overrideUrl");
                        System.exit(1);
                    }
                }
            }

            /* create server socket, later if DTLS */
            if (doDTLS == 0) {
                serverSocket = new ServerSocket(port);
            }

            InetAddress hostAddress = InetAddress.getLocalHost();
            System.out.println("Started server at " + hostAddress +
                    ", port " + port);

            /* wait for new client connections, then process */
            while (true) {

                System.out.println("\nwaiting for client connection...");
                if (doDTLS == 1) {
                    byte[] buf = new byte[1500];
                    d_serverSocket = new DatagramSocket(port);
                    DatagramPacket dp = new DatagramPacket(buf, buf.length);
                    d_serverSocket.setSoTimeout(0);
                    d_serverSocket.receive(dp);
                    d_serverSocket.connect(dp.getAddress(), dp.getPort());
                    System.out.println("client connection received from " + 
                            dp.getAddress() + " at port " + dp.getPort() +
                            "\n");
                } else {
                    clientSocket = serverSocket.accept();

                    /* get input and output streams */
                    outstream =
                        new DataOutputStream(clientSocket.getOutputStream());
                    instream =
                        new DataInputStream(clientSocket.getInputStream());
                    System.out.println("client connection received from " +
                            clientSocket.getInetAddress().getHostAddress() +
                            " at port " + clientSocket.getLocalPort() + "\n");
                }

                /* create SSL object */
                WolfSSLSession ssl = new WolfSSLSession(sslCtx);

                /* enable/load CRL functionality */
                ret = ssl.enableCRL(0);
                if (ret != WolfSSL.SSL_SUCCESS) {
                    System.out.println("failed to enable CRL, ret = "
                            + ret);
                    System.exit(1);
                }
                ret = ssl.loadCRL(crlPemDir, WolfSSL.SSL_FILETYPE_PEM,
                        (WolfSSL.CYASSL_CRL_MONITOR |
                         WolfSSL.CYASSL_CRL_START_MON));
                if (ret == WolfSSL.MONITOR_RUNNING_E) {
                    System.out.println("CRL monitor already running, " +
                            "continuing");
                } else if (ret != WolfSSL.SSL_SUCCESS) {
                    System.out.println("failed to start CRL monitor, ret = "
                            + ret);
                    System.exit(1);
                }
                MyMissingCRLCallback crlCb = new MyMissingCRLCallback();
                ret = ssl.setCRLCb(crlCb);
                if (ret != WolfSSL.SSL_SUCCESS) {
                    System.out.println("failed to set CRL callback, ret = "
                            + ret);
                    System.exit(1);
                }

                if (useIOCallbacks || (doDTLS == 1)) {
                    /* register I/O callbacks */
                    MyRecvCallback rcb = new MyRecvCallback();
                    MySendCallback scb = new MySendCallback();
                    MyIOCtx ioctx = new MyIOCtx(outstream, instream,
                            d_serverSocket, hostAddress, port);
                    sslCtx.setIORecv(rcb);
                    sslCtx.setIOSend(scb);
                    ssl.setIOReadCtx(ioctx);
                    ssl.setIOWriteCtx(ioctx);
                    System.out.println("Registered I/O callbacks");
                   
                    /* register DTLS cookie generation callback */ 
                    MyGenCookieCallback gccb = new MyGenCookieCallback();
                    MyGenCookieCtx gctx = new MyGenCookieCtx(
                            hostAddress, port);
                    sslCtx.setGenCookie(gccb);
                    ssl.setGenCookieCtx(gctx);
                    System.out.println("Registered DTLS cookie callback");

                } else {
                    /* if not using DTLS or I/O callbacks, pass Socket 
                     * fd to CyaSSL */
                     ret = ssl.setFd(clientSocket);

                    if (ret != WolfSSL.SSL_SUCCESS) {
                        System.out.println("Failed to set file descriptor");
                        return;
                    }
                }

                if (useAtomic == 1) {
                    /* register atomic record layer callbacks */
                    MyMacEncryptCallback mecb = new MyMacEncryptCallback();
                    MyDecryptVerifyCallback dvcb = 
                        new MyDecryptVerifyCallback();
                    MyAtomicEncCtx encCtx = new MyAtomicEncCtx();
                    MyAtomicDecCtx decCtx = new MyAtomicDecCtx();
                    sslCtx.setMacEncryptCb(mecb);
                    sslCtx.setDecryptVerifyCb(dvcb);
                    ssl.setMacEncryptCtx(encCtx);
                    ssl.setDecryptVerifyCtx(decCtx);
                }

                if (pkCallbacks == 1) {
                    /* register public key callbacks */
                    
                    /* ECC */
                    MyEccSignCallback eccSign = new MyEccSignCallback();
                    MyEccVerifyCallback eccVerify = new MyEccVerifyCallback();
                    MyEccSignCtx eccSignCtx = new MyEccSignCtx();
                    MyEccVerifyCtx eccVerifyCtx = new MyEccVerifyCtx();
                    sslCtx.setEccSignCb(eccSign);
                    sslCtx.setEccVerifyCb(eccVerify);
                    ssl.setEccSignCtx(eccSignCtx);
                    ssl.setEccVerifyCtx(eccVerifyCtx);

                    /* RSA */
                    MyRsaSignCallback rsaSign = new MyRsaSignCallback();
                    MyRsaVerifyCallback rsaVerify = new MyRsaVerifyCallback();
                    MyRsaEncCallback rsaEnc = new MyRsaEncCallback();
                    MyRsaDecCallback rsaDec = new MyRsaDecCallback();
                    MyRsaSignCtx rsaSignCtx = new MyRsaSignCtx();
                    MyRsaVerifyCtx rsaVerifyCtx = new MyRsaVerifyCtx();
                    MyRsaEncCtx rsaEncCtx = new MyRsaEncCtx();
                    MyRsaDecCtx rsaDecCtx = new MyRsaDecCtx();
                    sslCtx.setRsaSignCb(rsaSign);
                    sslCtx.setRsaVerifyCb(rsaVerify);
                    sslCtx.setRsaEncCb(rsaEnc);
                    sslCtx.setRsaDecCb(rsaDec);
                    ssl.setRsaSignCtx(rsaSignCtx);
                    ssl.setRsaVerifyCtx(rsaVerifyCtx);
                    ssl.setRsaEncCtx(rsaEncCtx);
                    ssl.setRsaDecCtx(rsaDecCtx);
                }

                ret = ssl.accept();
                if (ret != WolfSSL.SSL_SUCCESS) {
                    int err = ssl.getError(ret);
                    String errString = sslLib.getErrorString(err);
                    System.out.println("CyaSSL_accept failed. err = " + err +
                            ", " + errString);
                    System.exit(1);
                }

                /* show peer info */
                showPeer(ssl);

                /* read client response, and echo */
                insz = ssl.read(input, input.length);
                if (input.length > 0) {
                    String cliMsg = new String(input, 0, insz);
                    System.out.println("client says: " + cliMsg);
                } else {
                    System.out.println("read failed");
                }
                    
                ret = ssl.write(msg.getBytes(), msg.length());
                if (ret != msg.length()) {
                    System.out.println("ssl.write() failed");
                    System.exit(1);
                }

                ssl.shutdownSSL();
                ssl.freeSSL();

                if (doDTLS == 1) {
                    d_serverSocket.disconnect();
                    d_serverSocket.close();
                }
            }
        } catch (UnsatisfiedLinkError ule) {
            ule.printStackTrace();
        } catch (WolfSSLException wex) {
            wex.printStackTrace();
        } catch (CharacterCodingException cce) {
            cce.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

    } /* end run() */

    void showPeer(WolfSSLSession ssl) {

        String altname;
        long peerCrtPtr = ssl.getPeerCertificate();

        if (peerCrtPtr != 0) {

            System.out.println("issuer : " + ssl.getPeerX509Issuer(peerCrtPtr));
            System.out.println("subject : " + 
                    ssl.getPeerX509Subject(peerCrtPtr));

            while( (altname = ssl.getPeerX509AltName(peerCrtPtr)) != null)
                System.out.println("altname = " + altname);

        } else {
            System.out.println("peer has no cert!\n");
        }

        System.out.println("SSL version is " + ssl.getVersion());
        System.out.println("SSL cipher suite is " + ssl.cipherGetName());
    }

    void printUsage() {

        System.out.println("Java example server usage:");
        System.out.println("-?\t\tHelp, print this usage");
        System.out.println("-p <num>\tPort to connect to, default 11111");
        System.out.println("-v <num>\tSSL version [0-3], SSLv3(0) - " +
                "TLS1.2(3)), default 3");
        System.out.println("-l <str>\tCipher list");
        System.out.println("-c <file>\tCertificate file,\t\tdefault " +
                "../certs/client-cert.pem");
        System.out.println("-k <file>\tKey file,\t\t\tdefault " +
                "../certs/client-key.pem");
        System.out.println("-A <file>\tCertificate Authority file,\tdefault " +
                "../certs/client-cert.pem");
        System.out.println("-d\t\tDisable peer checks");
        System.out.println("-iocb\t\tEnable test I/O callbacks");
        System.out.println("-logtest\tEnable test logging callback");
        System.out.println("-U\t\tAtomic User Record Layer Callbacks");
        System.out.println("-P\t\tPublic Key Callbacks");
        System.exit(1);
    }

} /* end Client */

